;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Package: albert -*-

#|

DESC: tools/sds-asdf.lisp - tools for working with sysdefs
Copyright (c) 2003 - Stig Erik Sand

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

|#

(in-package :albert)

(defun mapappend (function list)
  (let ((f (coerce function 'function)))
    (loop for i in list append (funcall f i))))

(defun %get-default-directory ()
  "The default directory."
  #+allegro (excl:current-directory)
  #+clisp (ext:default-directory)
  #+cmu (ext:default-directory)
;;  #+sbcl (sb-ext:default-directory)
  #+cormanlisp (ccl:get-current-directory)
  #+lispworks (hcl:get-working-directory)
  #+lucid (lcl:working-directory)
  #+sbcl (truename ".")
  #-(or allegro sbcl clisp cmu cormanlisp lispworks lucid) (truename "."))

(defun get-system-pathname (system)
  "Returns the path to the system-def."

  (when (eq system nil)
    (return-from get-system-pathname nil))

  ;; improve this handling to let it magically determine what system it is
  #+asdf
  (etypecase system
    (asdf:system (asdf:component-pathname system))
    (string (get-system-pathname (asdf:find-system system)))
    (symbol (get-system-pathname (asdf:find-system system))))

  #-asdf
  (progn
    #+mk-defsystem
    (etypecase system
      (mk::component (mk::component-root-dir system :source))
      (string (get-system-pathname (mk:find-system system)))
      (symbol (get-system-pathname (mk:find-system system))))
    #-mk-defsystem
    nil))


#+asdf
(defgeneric all-asdf-components (component))

#+asdf
(defmethod all-asdf-components ((source-file asdf:cl-source-file))
  (list source-file))

#+asdf
(defmethod all-asdf-components ((source-file asdf:source-file))
  nil)

#+asdf
(defmethod all-asdf-components ((module asdf:module))
  (cons module (mapappend #'all-asdf-components (asdf:module-components module))))

#+asdf
(defmethod all-asdf-components ((module symbol))
  (all-asdf-components (asdf:find-system module)))


#+asdf
(defun get-system-files-asdf (system)
  "Returns a list of files for the system"
  (let* ((system (asdf:find-system system))
	 (sys-path  (asdf:component-pathname system))
	 (base-path (make-pathname
		     :directory (butlast (pathname-directory sys-path))
		     :defaults sys-path))
	 (comp-files (mapcar #'asdf:component-pathname (all-asdf-components system)))
	 (cur-dir (%get-default-directory))
	 (files (mapcar #'(lambda (x) (enough-namestring x cur-dir)) comp-files)))
    (remove-if-not #'pathname-name files)))

#+mk-defsystem
(defgeneric all-mkdefsys-components (component))

#+mk-defsystem
(defmethod all-mkdefsys-components ((component mk::component))
  (let ((components (mk::component-components component)))
    (cond (components 
	   (mapappend #'all-mkdefsys-components components))
          (t (list component)))))


#+mk-defsystem
(defmethod all-mkdefsys-components ((module symbol))
  (all-mkdefsys-components (mk:find-system module)))


#+mk-defsystem
(defun get-system-files-mkdefsys (system)
  "Returns a list of files for the system"
  (mapcar #'(lambda (c) (mk::component-full-pathname c :source))
          (all-mkdefsys-components system)))


#+asdf
(defun register-albert-settings-asdf (system)
  "Checks a system for info and registers it as albert-settings."

  (unless system
    (error "No system provided for registering settings."))
    
  (unless (typep system 'asdf::system)
    (setf system (asdf:find-system system)))

  (unless (typep system 'asdf::system)
    (error "Unable to find asdf:system specified as: ~s" system))
  
  ;;(warn "&&& Registering stuff from ~s" system)

  (ignore-errors
    (when-bind (name (asdf::component-name system))
      (setf (albert-setting '("system" "name")) name)))

  (ignore-errors
    (when-bind (path (get-system-pathname system))
      (setf (albert-setting '("system" "directory")) path)))
  
  (ignore-errors
    (when-bind (author (asdf::system-author system))
      (setf (albert-setting '("system" "author" "name")) author)))

  (ignore-errors
    (when-bind (maint (asdf::system-maintainer system))
      (setf (albert-setting '("system" "maintainer" "name")) maint)))

  (ignore-errors
    (when-bind (author (asdf::system-licence system))
      (setf (albert-setting '("system" "licence" "name")) author)))

  (ignore-errors
    (when-bind (version (asdf::component-version system))
      (setf (albert-setting '("system" "version" "number")) version)))

  (ignore-errors
    (when-bind (desc (asdf::system-description system))
      (setf (albert-setting '("system" "description")) desc)))

  (flet ((to-string (arg)
	   (typecase arg
	     (string arg)
	     (symbol (string-downcase (string arg))))))
	     
  (ignore-errors
    ;; assumes it's an alist
    (let ((props (asdf::component-properties system)))
      (dolist (i props)
	
	(cond ((consp i)

	       (cond ((consp (car i)) ;; some are lists

		      (cond ((equal (car i) '("system" "author" "email"))
			     (setf (albert-setting (car i)) (cdr i)))
			    ((equal (car i) '("system" "date"))
			     (setf (albert-setting (car i)) (cdr i)))
			    ((equal (car i) '("system" "directory"))
			     (setf (albert-setting (car i)) (cdr i)))
			    ((equalp (mapcar #'to-string (car i)) '("albert" "formats"))
			     (setf (albert-setting '("albert" "presentation" "formats")) (cdr i)))
			    ((equalp (mapcar #'to-string (car i)) '("albert" "output-dir"))
			     (setf (albert-setting '("albert" "presentation" "output-dir")) (cdr i)))
			    ((equalp (mapcar #'to-string (car i)) '("albert" "docbook" "template"))
			     (setf (albert-setting '("albert" "docbook" "generate")) (cdr i)))
			    ((string-equal (string (first (car i))) "albert") ;; if albert is first in the list
			     (setf (albert-setting (mapcar #'to-string (car i))) (cdr i)))
			    (t
			     nil)))

		     ((or (symbolp (car i)) (stringp (car i)))
		      (cond ((string-equal (string (car i)) "author-email")
			     (setf (albert-setting '("system" "author" "email")) (cdr i)))
			    ((string-equal (string (car i)) "date")
			     (setf (albert-setting '("system" "date")) (cdr i)))
			    ((string-equal (string (car i)) "licencefile")
			     (setf (albert-setting '("system" "licencefile")) (cdr i)))
			    (t
			     (warn "Albert-ASDF: Checking, but not handling: ~s" (car i)))))

		     
		     ))
	      (t
	       (warn "Odd format for property: ~s" i)))
	)))

  t))

(defun register-albert-settings (system)
  "Checks a system for info and registers it as albert-settings."
  #+asdf
  (register-albert-settings-asdf system)
  #-asdf
  (warn "Only ASDF supported for REGISTER-ALBERT-SETTINGS yet."))


(defun get-system-files (system)
  "Returns a list of files for the system"
  #+asdf
  (get-system-files-asdf system)
  #-asdf
  (progn
    #+mk-defsystem
    (get-system-files-mkdefsys system)
    #-mk-defsystem
    (warn "Only ASDF and MKDEFSYS supported for GET-SYSTEM-FILES yet.")))

(defun show-files (system)
  (let ((files (get-system-files system)))
    (dolist (i files)
      (format t "~&~s~%" i))))

(defun document-files (files &key (use-clean-readtable t))
  "Documents the list of FILES."

  (unless files
    (warn "No files submitted to DOCUMENT-FILES")
    (return-from document-files nil))
  
  (let ((retval nil)
	(apispec-base:*verbose-operation* (albert-setting '("albert" "verbose")))
	(apispec-base:*submarine-quiet* (albert-setting '("albert" "submarine-quiet")))
	(spres-impl:*rule-trace* (albert-setting '("albert" "presentation" "trace-rules")))
	(spres-impl:*clhs-root* (albert-setting '("hyperspec" "root")))
	(cl:*load-verbose* (albert-setting '("albert" "verbose")))
	(temporary-file nil)
	(csf-tree nil)
	(sdoc-tree nil)
	)

    (when-verbose
	(albert-info "Will now analyse files in given order:~%   ~s" files))

    (when (albert-setting '("albert" "use-temporary-files"))
      (setf temporary-file (albert-setting '("albert" "lisp2csf" "outfile"))))
    
    (setf csf-tree (lisp2csf:analyse-files files :out-file temporary-file
					   :use-clean-readtable use-clean-readtable))
    
    (unless (consp csf-tree)
      (warn "LISP2CSF seemed to fail (no returned value), exiting")
      (return-from document-files nil))
    
    ;;(warn "RETVAL is ~s" csf-tree)
    
    (when (albert-setting '("albert" "use-temporary-files"))
      (setf csf-tree (csf:parse-csf-file temporary-file)) ;; last file
      (unless (consp csf-tree)
	(warn "Unable to load temporary CSF-file: ~s" temporary-file)
	(return-from document-files nil)))

    (setf temporary-file nil)
    
    (when (albert-setting '("albert" "use-temporary-files"))
      (setf temporary-file (albert-setting '("albert" "csf2sdoc" "outfile"))))
    
    
    (setf sdoc-tree (albert:convert-csf-to-sdoc csf-tree :out-file temporary-file ))

    (unless (consp sdoc-tree)
      (warn "CSF2SDOC seemed to fail (no returned value), exiting")
      (return-from document-files nil))
    
    ;;(warn "RETVAL is ~s" sdoc-tree)

    (when (albert-setting '("albert" "use-temporary-files"))
      (setf sdoc-tree (sdoc:parse-sdoc-file temporary-file)) ;; last file
      (unless (consp sdoc-tree)
	(warn "Unable to load temporary SDOC-file: ~s" temporary-file)
	(return-from document-files nil)))
    
    
    (setf retval (spres:present-sdoc sdoc-tree))
    
    ;;(warn "RETVAL is ~s" retval)
    
    retval))

(defun document-systems (&rest systems)
  "Produces a DocBook XML file which contains markup that represents
packages, methods, classes, functions, etc. extracted from 'systems'."
  (mapc #'register-albert-settings (reverse systems))
  (document-files (flatten (mapcar #'(lambda (system)
                                       (get-system-files system)) systems))))
